import logging
import contextlib

from typing import Union, List
from json.decoder import JSONDecodeError

from msgraph import core
from azure import identity

from ppa import ui


from . import exceptions


logging.getLogger("azure").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


EXCEPTION_MAP = {
    "does not comply with password complexity": lambda action: (
        exceptions.InvalidPassword,
        f"Cannot {action} as the supplied password does not meet complexity requirements.",
    ),
    "passwordProfile.password has an invalid value": lambda action: (
        exceptions.InvalidPassword,
        f"Cannot {action} as the supplied password is invalid.",
    ),
    "insufficient privileges": lambda action: (
        exceptions.InsufficientPrivileges,
        f"Cannot {action} as PPA lacks the relevant API permissions.",
    ),
    "already exists": lambda action: (
        exceptions.AlreadyExists,
        f"Cannot {action} as a user already exists with that user principal name.",
    ),
    "does not exist": lambda action: (exceptions.DoesNotExist, f"Cannot {action} as it does not exist."),
    "one or more added object references already exist": lambda action: (
        exceptions.AlreadyExists,
        f"Cannot {action} as it already exists in the group.",
    ),
    "one or more removed object references do not exist": lambda action: (
        exceptions.DoesNotExist,
        f"Cannot {action} as it is not in the group.",
    ),
    "domain portion of the userPrincipalName property": lambda action: (
        exceptions.InvalidParameter,
        f"Cannot {action} as the domain portion of the supplied user principal name is invalid.",
    ),
    "does not have any available licenses": lambda action: (
        exceptions.SubscriptionError,
        f"Cannot {action} as the subscription has no remaining licences.",
    ),
}


def _generate_exception_message(status_code: int, request_type: str, url: str, response_body: Union[dict, str]) -> str:
    """
    {
        'error': {
            'code': 'Request_BadRequest',
            'message': 'Another object with the same value for property userPrincipalName already exists.',
            'innerError': {
                'date': '2021-09-28T08:22:40',
                'request-id': '9011063c-2a88-4188-adf4-82ace56f86bc',
                'client-request-id': '5b16cf07-09bf-41bf-bf3b-9bd234ce135b'
            }
        }
    }
    """
    return (
        f"({status_code}) Failed {request_type} request to {url}: "
        f"{response_body.get('error', {}).get('message') or response_body if isinstance(response_body, dict) else response_body}"
    )


def _get_matching_exception(exception_message):
    for key in EXCEPTION_MAP:
        if key.lower() in exception_message.lower():
            return EXCEPTION_MAP[key]


@contextlib.contextmanager
def _exception_handler(action, ignored_exceptions=None):
    try:
        yield
    except exceptions.RequestError as e:
        if exception_generator := _get_matching_exception(str(e)):
            exception, message = exception_generator(action)
            if ignored_exceptions is not None and exception in ignored_exceptions:
                logger.info(f"ignoring reraised exception {exception}")
                return
            ui.output_error(message)
            raise exception(message)
        raise e


def graph_api(func):
    def _execute(client, *args, **kwargs):
        # Only check keys & create a GraphClient if it isn't one already (supports one decorated function calling another)
        if isinstance(client, dict):
            if missing_keys := [k for k in {"tenant_id", "client_id", "client_secret"} if k not in client]:
                message = "Supplied graph client is missing the following required keys: {}".format(", ".join(missing_keys))
                ui.output_error(message)
                raise exceptions.BadClientConfiguration(message)
            return func(core.GraphClient(credential=identity.ClientSecretCredential(**client)), *args, **kwargs)
        return func(client, *args, **kwargs)

    return _execute


def get_response_body(response) -> Union[dict, str]:
    try:
        return response.json()
    except JSONDecodeError:
        # Sometimes a response to a failed request won't have a JSON body.
        return response.text or ""


def assert_status_code(response, response_body, valid_status_codes: Union[List[int], int]):
    if response.status_code not in ([valid_status_codes] if not isinstance(valid_status_codes, list) else valid_status_codes):
        raise exceptions.RequestError(
            _generate_exception_message(response.status_code, response.request.method, response.request.url, response_body)
        )


def get(client, url):
    results = []

    for i in range(0, 50):
        logger.info(f"sending GET request to endpoint {url}")
        response = client.get(url)
        response_body = get_response_body(response)
        assert_status_code(response, response_body, 200)

        # If assert_status_code hasn't raised an exception, we can safely assume the response body is JSON.
        if "value" in response_body:
            # The data returned by GET requests is usually nested inside the 'value' key of the
            # response, but it isn't when you get a single object (like users.by_principal_name).
            results.extend(response_body["value"])
        else:
            # If there is no 'value' key then a single object has been retrieved, & it can be returned
            # without looking for a next page.
            logger.debug("No 'value' dictionary was found in the Graph API response, returning the full response instead")
            return response_body

        if next_page_link := response_body.get("@odata.nextLink"):
            logger.info("there is another results page...")
            url = next_page_link
        else:
            break
    else:
        ui.output_error("Over 50 pages of results were received, only the first 50 will be returned.")

    logger.info(f"collected {len(results)} results")
    return results


def post(client, url, json):
    logger.info(f"sending POST request to endpoint {url}")
    response = client.post(url, json=json)
    response_body = get_response_body(response)
    assert_status_code(response, response_body, [200, 201, 204])
    return response_body


def patch(client, url, json):
    logger.info(f"sending PATCH request to endpoint {url}")
    response = client.patch(url, json=json)
    response_body = get_response_body(response)
    assert_status_code(response, response_body, 204)
    return response_body


def delete(client, url, json=None):
    logger.info(f"sending DELETE request to endpoint {url}")
    response = client.delete(url, json=json)
    response_body = get_response_body(response)
    assert_status_code(response, response_body, 204)
    return response_body
